use std::any::TypeId;
use std::cell::RefCell;
use std::collections::HashMap;
use std::ops::Deref;
use std::ops::DerefMut;
use std::{borrow::BorrowMut, rc::Rc};

use crate::io::{FleckIO, Packet};

mod event;
mod lowlevel;
mod nodes;
mod priority;

pub use priority::{Never, ServicePriority as BetweenPriority};

pub use event::ChannelSpec;

use self::priority::AbstractServicePriority;

pub mod order {
    pub use super::priority::Never;

    pub struct FirstTag;
    pub type First = super::priority::Fixpoint<FirstTag>;

    pub struct PreprocessingTag;
    pub type Preprocessing = super::priority::Fixpoint<PreprocessingTag>;

    pub struct ProcessingTag;
    pub type Processing = super::priority::Fixpoint<ProcessingTag>;

    pub struct PostprocessingTag;
    pub type Postprocessing = super::priority::Fixpoint<PostprocessingTag>;

    pub struct LastTag;
    pub type Last = super::priority::Fixpoint<FirstTag>;
}

pub mod core {
    pub(crate) use super::lowlevel::{LocalDiscovery, ParsePacket, SendPacket};

    pub use super::nodes::Nodes;

    pub mod channel_tags {
        #[derive(Default)]
        pub struct ReceivePacketTag {}
        #[derive(Default)]
        pub struct MinorTickTag {}
        #[derive(Default)]
        pub struct MajorTickTag {}
    }

    pub use super::lowlevel::SendPacketChannel;
    pub type ReceivePacketChannel = (channel_tags::ReceivePacketTag, crate::io::Packet);
    pub type MinorTickChannel = (channel_tags::MinorTickTag, ());
    pub type MajorTickChannel = (channel_tags::MajorTickTag, ());
}

pub type EventRoot = event::EventRoot<crate::Fleck>;

#[allow(unused_variables)]
pub trait Service: std::any::Any + crate::helper::AsAny {
    /*fn process_incoming(&mut self, api: &crate::Fleck, packet: &mut Packet) {}
    fn process_outgoing(&mut self, api: &crate::Fleck, packet: &mut Packet) {}
    fn process_minor_tick(&mut self, api: &crate::Fleck) {}
    fn process_major_tick(&mut self, api: &crate::Fleck) {}*/

    fn register_channels(&self, eroot: &mut EventRoot) {}
    fn register_msgs(&self, registry: &mut crate::msg::MessageRegistry) {}
}

trait ServiceExt {
    fn build() -> Self;
}

impl<T: Service + Default> ServiceExt for T {
    fn build() -> Self {
        Default::default()
    }
}

#[derive(Default)]
pub struct ServiceStack {
    services: Vec<Box<dyn Service>>,

    service_map: HashMap<TypeId, Box<dyn Service>>,

    eroot: EventRoot,
    message_registry: crate::msg::MessageRegistry,
}

impl ServiceStack {
    pub fn give_service<S: 'static + Default>(&mut self)
    where
        Rc<S>: Service,
    {
        let srv = Rc::new(S::default());

        srv.register_msgs(&mut self.message_registry);
        srv.register_channels(&mut self.eroot);

        self.service_map
            .insert(TypeId::of::<S>(), Box::new(srv.clone()));
        self.services.push(Box::new(srv));
    }

    pub fn with_service<S: 'static, R, F: FnOnce(&S) -> R>(&self, f: F) -> R
    where
        Rc<S>: Service,
    {
        let id = TypeId::of::<S>();

        self.service_map
            .get(&id)
            .expect("asked for service that doesn't exist!")
            .deref()
            .as_any()
            .downcast_ref()
            .map(f)
            .unwrap()
    }

    pub fn event_root(&self) -> &EventRoot {
        &self.eroot
    }
}

impl ServiceStack {
    pub(crate) fn process_incoming(&self, api: &crate::Fleck, packet: Packet) {
        self.eroot
            .priority_channel::<core::ReceivePacketChannel>()
            .queue(packet);
        self.eroot.fire_all(api);
    }

    pub(crate) fn process_minor_tick(&self, api: &crate::Fleck) {
        self.eroot.channel::<core::MinorTickChannel>().queue(());
        self.eroot.fire_all(api);
    }

    pub(crate) fn process_major_tick(&self, api: &crate::Fleck) {
        self.eroot.channel::<core::MajorTickChannel>().queue(());
        self.eroot.fire_all(api);
    }

    pub(crate) fn create_io_channels(&mut self) {
        self.eroot
            .create_priority_channel::<core::ReceivePacketChannel>();
        self.eroot.create_channel::<core::MinorTickChannel>();
        self.eroot.create_channel::<core::MajorTickChannel>();
    }
}
